//! Authenticated endpoints for the [content repository].
//!
//! [content repository]: https://spec.matrix.org/v1.18/client-server-api/#content-repository

pub mod get_content;
pub mod get_content_as_filename;
pub mod get_content_thumbnail;
pub mod get_media_config;
pub mod get_media_preview;
