//! Endpoints for third party lookups

pub mod get_location_for_protocol;
pub mod get_location_for_room_alias;
pub mod get_protocol;
pub mod get_protocols;
pub mod get_user_for_protocol;
pub mod get_user_for_user_id;
