//! Endpoints for [previewing a room] without joining it, aka peeking.
//!
//! [previewing a room]: https://spec.matrix.org/v1.18/client-server-api/#room-previews

pub mod get_current_state;
pub mod listen_to_new_events;
