//! Methods and types for (de)serialization.

pub mod as_str_as_ref_str;
pub mod debug_as_ref_str;
pub mod deserialize_from_cow_str;
pub mod display_as_ref_str;
pub mod enum_as_ref_str;
pub mod enum_from_string;
pub mod eq_as_ref_str;
pub mod ord_as_ref_str;
pub mod serialize_as_ref_str;
mod util;
