//! Endpoints to create an association between a Matrix ID and an email ID on the identity server.

pub mod create_email_validation_session;
pub mod validate_email;
pub mod validate_email_by_end_user;
