//! Endpoints to create association between a phone number and a Matrix ID on the identity server.

pub mod create_msisdn_validation_session;
pub mod validate_msisdn;
pub mod validate_msisdn_by_phone_number;
